SystemOrganization addCategory: #'Xpf-Multipoint-Core'!

!HandMorph methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/31/2009 14:40'!
isXpfMultiPointerHandMorph

	^ false! !

!HandMorph methodsFor: '*xpf-multipoint-core' stamp: 'SR 8/27/2009 20:48'!
lastEventBuffer
	^ lastEventBuffer! !

!HandMorph methodsFor: '*xpf-multipoint-core' stamp: 'SR 8/27/2009 20:48'!
lastEventBuffer: anObject
	lastEventBuffer := anObject! !

!HandMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 8/27/2009 20:27'!
processEvents
	"Process user input events from the local input devices."
	| evtBuf hadAny shouldProcessEvents |

	shouldProcessEvents := [MultiPointerManager instance moving not] ifError: [true].
	
	shouldProcessEvents ifTrue: [
		ActiveEvent ifNotNil: [
			"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
	
	hadAny := false.
	
	[(evtBuf := Sensor nextEvent) notNil]
		whileTrue: [ | evt type |
			type := evtBuf first.
			evt := nil.
			
			shouldProcessEvents 
				ifTrue: [
					"normal behaviour"
					type = EventTypeMouse ifTrue: [evt := self generateMouseEvent: evtBuf].
					type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf].
					type = EventTypeDragDropFiles ifTrue: [evt := self generateDropFilesEvent: evtBuf].
				
					evt ifNotNil: [self handleEvent: evt. hadAny := true]]].
	
	
	shouldProcessEvents ifTrue: [
		"note: if we come here we didn't have any mouse events"
		mouseClickState ifNotNil: [
			"No mouse events during this cycle. Make sure click states time out accordingly"
			mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
		hadAny ifFalse: ["No pending events. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
! !

HandMorph subclass: #MultiPointerHandMorph
	instanceVariableNames: 'queue mouseIndex keyboardIndex normalCursor handColor movedTimestamp lastPosition lastModifiers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'mig 6/4/2009 17:27'!
cursorSize

	^ 35! !

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'SR 5/31/2009 14:35'!
displayTimeout

	^ 10000! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 15:51'!
keyboardIcon

	^ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self keyboardIconContents readStream)! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 17:29'!
keyboardIconContents

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJ
bWFnZVJlYWR5ccllPAAAAdxJREFUeNrEU0tLG1EU/iadgggVMUJomhQxPiuU+CjURTZWEJFuuhLd
uelfCP6AbsWlP6BQcVtaqC/sxkWhDUUIjSUSDAaNGJLMZJI799Fzb2KEFgriwgPffOfemXu+c8+Z
YymlcBcL4I5mXzuZTOat7/trhAZZrV6vVz3PKxJStVrtgrhMfEzYTiaT/J8A9ME6HQQhmUgkNm6d
geM4D0n1MwXavNUVvp6IU+KI63qgAAvdz+ZKO78a4EISBHwuwH0BRuz7es0Ns6af1xlEQnYVGJ0y
EZVyoBsjNeihuySIpdR7D8i3yLeJgYP0VaTdhYHwI+QvXAw+6ULuvIqRaBeyZxWM9XXjOF9GfKAH
6VwJL4Z7zZ7OkDEOa/83UyHbQb7oGkX9QitygjDX0FBtjseCmBgKYu9HAbs/iwhwLs3BmfHHmJ0M
010lFl5G0WACbxJ9qDOJpVcxWkuszA+Zw9p0QF0P69ORq8IdHk4KFXCulW9UdQZ+mykzcZPZ8/4g
tlMF2KbKpPp6+ik+7GWxONP/37a9e5/C6nIcHw9PjZC19e1KRTsZRKvqUkpT4TYr1dpXxteTo2ST
v3w/g91g/DLtqt6/+tv0W/+AyVKIZoEJktAawUvr3qfxjwADAPOBecF+sUNmAAAAAElFTkSuQmCC'! !

!MultiPointerHandMorph class methodsFor: 'misc' stamp: 'mig 5/27/2009 12:14'!
scaledCursor: size colored: aColor

	| form canvas leftPoint rightPoint barOffset borderWidth baseOffset |
	form := Form extent: size@size depth: 32.
	canvas := FormCanvas on: form.
	borderWidth := ((size / 20) max: 1) rounded.
	baseOffset := borderWidth@borderWidth.
	
	rightPoint := Point r: size*0.75 degrees: 45.
	leftPoint := Point r: size*0.75 degrees: 90.
	barOffset := Point r: size*0.2 degrees: 62.5.
	
	canvas
		drawPolygon: {
			baseOffset.
			baseOffset + rightPoint.
			baseOffset + (leftPoint * 0.33) + (rightPoint * 0.67).
			baseOffset + (leftPoint * 0.33) + (rightPoint * 0.67) + barOffset.
			baseOffset + (leftPoint * 0.67) + (rightPoint * 0.33) + barOffset.
			baseOffset + (leftPoint * 0.67) + (rightPoint * 0.33).
			baseOffset + leftPoint} 
		fillStyle: aColor
		borderWidth: borderWidth
		borderColor: Color black.
	
	^ (form asCursorForm) offset: baseOffset negated; yourself! !

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'mig 6/4/2009 17:39'!
waveSteps

	^ (1.0 to: 3.0 by: 0.3), (3.0 to: 1.0 by: -0.5)! !

!MultiPointerHandMorph methodsFor: 'updating' stamp: 'mig 5/19/2009 20:53'!
changed

	bounds := bounds origin extent: self currentCursor extent.
	super changed.! !

!MultiPointerHandMorph methodsFor: 'cursor' stamp: 'mig 5/18/2009 14:18'!
currentCursor

	^ {self temporaryCursor. self normalCursor. NormalCursor} detect: [:cursor| cursor notNil]! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 8/27/2009 14:02'!
drawOn: aCanvas 
	"Draw the hand itself (i.e., the cursor)."
		
	self hasMovedInLastTime ifFalse: [^ self].
	aCanvas 
		paintImage: self currentCursor
		at: bounds topLeft.! !

!MultiPointerHandMorph methodsFor: 'as yet unclassified' stamp: 'SR 8/27/2009 18:16'!
enqueueEvent: evtBuf

	self queue nextPut: evtBuf.! !

!MultiPointerHandMorph methodsFor: 'as yet unclassified' stamp: 'SR 9/1/2009 15:07'!
generateKeyboardEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"

	| buttons modifiers type pressType stamp char |
	stamp := Time millisecondClockValue.
	pressType := evtBuf fourth.
	pressType = EventKeyDown ifTrue: [type := #keyDown].
	pressType = EventKeyUp ifTrue: [type := #keyUp].
	pressType = EventKeyChar ifTrue: [type := #keystroke].
	modifiers := evtBuf fifth.
	buttons := modifiers bitShift: 3.
	char := self keyboardInterpreter nextCharFrom: Sensor firstEvt: evtBuf.
	
	(type = #keystroke
		and: [(buttons anyMask: 16) 
		and: [char asciiValue = 30 or: [char asciiValue = 31]]])
		ifTrue: [^ MouseWheelEvent new 
					setType: #mouseWheel
					position: lastMouseEvent cursorPoint
					direction: (char asciiValue = 30 ifTrue: [#up] ifFalse: [#down])
					buttons: buttons
					hand: self
					stamp: stamp].
	
	^ KeyboardEvent new
		setType: type
		buttons: buttons
		position: self position
		keyValue: char asciiValue
		hand: self
		stamp: stamp.! !

!MultiPointerHandMorph methodsFor: 'as yet unclassified' stamp: 'SR 8/27/2009 20:50'!
generateMouseEvent: evtBuf 
	"Generate the appropriate mouse event for the given raw event buffer"

	| deltaPosition buttons type stamp oldButtons trail |
	stamp := Time millisecondClockValue. "second"
	deltaPosition := evtBuf third @ evtBuf fourth.
	buttons := evtBuf fifth bitOr: (self lastModifiers bitShift: 3).
	
	type := (buttons = 0) 
		ifTrue: [self lastEventBuffer fifth = 0
			ifTrue: [#mouseMove]
			ifFalse: [#mouseUp]]
		ifFalse: [self lastEventBuffer fifth = 0 
			ifTrue: [#mouseDown]
			ifFalse: [#mouseMove]].
	
	oldButtons := self lastEventBuffer fifth bitOr: (self lastModifiers bitShift: 3).
	
	self lastEventBuffer: evtBuf.
	self lastPosition: ((0@0 max: (self lastPosition + deltaPosition)) min: World screenRectangle bottomRight).
	
	type == #mouseMove ifTrue: [
		trail := self mouseTrailFrom: evtBuf.
		^ MouseMoveEvent new 
			setType: type
			startPoint: self position
			endPoint: trail last
			trail: trail
			buttons: buttons
			hand: self
			stamp: stamp].
	
	^ MouseButtonEvent new 
		setType: type
		position: self lastPosition
		which: (oldButtons bitXor: buttons)
		buttons: buttons
		hand: self
		stamp: stamp! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:15'!
handColor

	^ handColor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 6/4/2009 17:28'!
handColor: aColor

	handColor := aColor.
	self normalCursor: (self class scaledCursor: self class cursorSize colored: aColor).! !

!MultiPointerHandMorph methodsFor: 'events-processing' stamp: 'SR 8/27/2009 18:06'!
handleEvent: anEvent
	
	(anEvent isMouse and: [anEvent isMouseDown]) ifTrue: [
		(anEvent redButtonPressed and: [anEvent yellowButtonPressed]) ifTrue: [self wave]].
	(anEvent isMouse and: [anEvent isMouseMove]) ifTrue: [
		self resetTimeout].
	^ super handleEvent: anEvent! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:56'!
hasMovedInLastTime

	^ self movedTimestamp notNil 
		and: [Time millisecondClockValue - self movedTimestamp < self class displayTimeout]! !

!MultiPointerHandMorph methodsFor: 'initialization' stamp: 'SR 8/27/2009 20:45'!
initialize

	super initialize.
	self
		queue: SharedQueue new;
		lastPosition: 0@0;
		lastModifiers: 0.! !

!MultiPointerHandMorph methodsFor: 'testing' stamp: 'SR 5/31/2009 14:39'!
isXpfMultiPointerHandMorph

	^ true! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 14:02'!
keyboardIndex

	^ keyboardIndex! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 14:02'!
keyboardIndex: anObject

	keyboardIndex := anObject! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 20:45'!
lastModifiers

	^ lastModifiers! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 20:45'!
lastModifiers: anObject

	lastModifiers := anObject! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 20:43'!
lastPosition

	^ lastPosition! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 20:43'!
lastPosition: anObject

	lastPosition := anObject! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 14:03'!
mouseIndex

	^ mouseIndex! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 8/27/2009 14:03'!
mouseIndex: anObject

	mouseIndex := anObject! !

!MultiPointerHandMorph methodsFor: 'as yet unclassified' stamp: 'SR 8/27/2009 22:04'!
mouseTrailFrom: currentBuf 
	"Current event, a mouse event buffer, is about to be processed.  If there are other similar mouse events queued up, then drop them from the queue, and report the positions inbetween."

	| nextEvent trail |
	trail := WriteStream on: (Array new: 1).
	trail nextPut: self lastPosition.
	[(nextEvent := self queue peek) isNil] whileFalse: 
			[nextEvent first = currentBuf first 
				ifFalse: [^trail contents	"different event type"].
			nextEvent fifth = currentBuf fifth 
				ifFalse: [^trail contents	"buttons changed"].
			nextEvent sixth = currentBuf sixth 
				ifFalse: [^trail contents	"modifiers changed"].
			"nextEvent is similar.  Remove it from the queue, and check the next."
			nextEvent := self queue next.
			self lastPosition: ((0@0 max: (self lastPosition + (nextEvent third @ nextEvent fourth))) min: World screenRectangle bottomRight).
			trail nextPut: self lastPosition].
	^trail contents! !

!MultiPointerHandMorph methodsFor: 'updating' stamp: 'mig 6/3/2009 10:07'!
moveToPosition: aPosition
	"Issue a mouse move event to make the receiver appear at the given position"
	self lastPosition: aPosition.
	self handleEvent: (MouseMoveEvent new
		setType: #mouseMove 
		startPoint: self position 
		endPoint: aPosition 
		trail: (Array with: self position with: aPosition)
		buttons: 0
		hand: self
		stamp:  Time millisecondClockValue)! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp

	^ movedTimestamp! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp: anObject

	movedTimestamp := anObject.! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 4/24/2009 13:39'!
needsToBeDrawn

	^ true! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 10:41'!
normalCursor

	^ normalCursor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:53'!
normalCursor: anObject

	normalCursor := anObject.
	self changed.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'SR 8/27/2009 18:24'!
processEvents
	| hadAny |
	
	ActiveEvent ifNotNil: [
		"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent].

	hadAny := false.
	[self queue isEmpty] whileFalse: [ | evt evtBuf type |
		evtBuf := self queue next.
		evt := nil.	"for unknown event types"
	
		type := evtBuf first.
		type = EventTypeMouse ifTrue: [evt := self generateMouseEvent: evtBuf].
		type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf].
		
		"All other events are ignored"
		evt ifNotNil: [
			"Finally, handle it"
			self handleEvent: evt.
			hadAny := true.

			"For better user feedback, return immediately after a mouse event has been processed."
			(evt isMouse and: [evt isMouseWheel not]) ifTrue: [^self]]].
	
	"note: if we come here we didn't have any mouse events"
	mouseClickState ifNotNil: [
		"No mouse events during this cycle. Make sure click states time out accordingly"
		mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
	hadAny ifFalse: [
		"No pending events. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent]
! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
queue

	^ queue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:21'!
queue: anObject

	queue := anObject.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'SR 8/27/2009 14:03'!
resetTimeout

	| hasMovedBefore |
	hasMovedBefore := self hasMovedInLastTime.
	self movedTimestamp: Time millisecondClockValue.
	(hasMovedBefore not and: [self hasMovedInLastTime]) ifTrue: [self wave].! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'mig 6/4/2009 17:37'!
wave

	[
		self class waveSteps do: [:factor |
			self showTemporaryCursor: 
				(self class 
					scaledCursor: self class cursorSize * factor 
					colored: self handColor).
			(Delay forMilliseconds: 20) wait].
		
		self showTemporaryCursor: nil.
	] fork.! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/20/2009 19:46'!
blinkPrevParen
	| openDelimiter closeDelimiter level string here hereChar |
	string _ paragraph text string.
	here _ self startBlock stringIndex.
	openDelimiter _ sensor keyboardPeek.
	closeDelimiter _ '([{' at: (')]}' indexOf: openDelimiter).
	level _ 1.
	[level > 0 and: [here > 2]]
		whileTrue:
			[hereChar _ string at: (here _ here - 1).
			hereChar = closeDelimiter
				ifTrue:
					[level _ level - 1.
					level = 0
						ifTrue: [^ self blinkParenAt: here]]
				ifFalse:
					[hereChar = openDelimiter
						ifTrue: [level _ level + 1]]].! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/20/2009 20:55'!
markBlock

	^ (self markBlockForHand: ActiveHand) ifNil: [paragraph defaultCharacterBlock]! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 14:39'!
markBlock: aCharacterBlock

	stopBlock ifNil: [
		stopBlock := Dictionary new].
	stopBlock isDictionary ifFalse: [
		stopBlock := Dictionary new].
	stopBlock at: ActiveHand put: aCharacterBlock.
! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/20/2009 20:45'!
markBlockForHand: aHand

	^ stopBlock isDictionary
		ifTrue: [stopBlock at: aHand ifAbsent: [nil]]
		ifFalse: [stopBlock]! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/20/2009 20:55'!
pointBlock

	^ (self pointBlockForHand: ActiveHand) ifNil: [paragraph defaultCharacterBlock]! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 14:39'!
pointBlock: aCharacterBlock

	startBlock ifNil: [
		startBlock := Dictionary new].
	startBlock isDictionary ifFalse: [
		startBlock := Dictionary new].
	startBlock at: ActiveHand put: aCharacterBlock.! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/20/2009 20:45'!
pointBlockForHand: aHand

	^ startBlock isDictionary
		ifTrue: [startBlock at: aHand ifAbsent: [nil]]
		ifFalse: [startBlock]! !

!TextMorphEditor methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/7/2009 13:06'!
storeSelectionInParagraph

	World handsDo: [:hand | | currentMarkBlock currentPointBlock |
		currentMarkBlock := self markBlockForHand: hand.
		currentPointBlock := self pointBlockForHand: hand.
		
		(currentPointBlock isNil or: [currentMarkBlock isNil]) ifFalse: [
			(currentPointBlock stringIndex <= currentMarkBlock stringIndex)
				ifTrue: [paragraph selectionStart: currentPointBlock selectionStop: currentMarkBlock forHand: hand]
				ifFalse: [paragraph selectionStart: currentMarkBlock  selectionStop: currentPointBlock forHand: hand]]]! !

!SystemWindow class methodsFor: '*xpf-multipoint-core' stamp: 'mig 5/28/2009 11:36'!
cascadeWindows

	(SystemWindow windowsIn: World satisfying: [:window | window isCollapsed not]) withIndexDo: [:window :i |
		window position: (30*i) asPoint.
		window extent: ((window bottomRight adhereTo: Display fullBoundingBox) - window topLeft).
		window comeToFront].! !

!SystemWindow class methodsFor: '*xpf-multipoint-core' stamp: 'mig 5/28/2009 11:46'!
orderWindowsHorizontally

	| windowList windowExtent |

	windowList := SystemWindow windowsIn: World satisfying: [:window | window isCollapsed not].
	windowExtent := Display fullBoundingBox extent // (windowList size @ 1).
	
	windowList withIndexDo: [:window :i |
		window 
			position: (i-1 * windowExtent x) @ 0;
			extent: windowExtent;
			comeToFront].! !

!SystemWindow class methodsFor: '*xpf-multipoint-core' stamp: 'mig 5/28/2009 11:48'!
orderWindowsVertically

	| windowList windowExtent |

	windowList := SystemWindow windowsIn: World satisfying: [:window | window isCollapsed not].
	windowExtent := Display fullBoundingBox extent // (1 @ windowList size).
	
	windowList withIndexDo: [:window :i |
		window 
			position: 0 @ (i-1 * windowExtent y);
			extent: windowExtent;
			comeToFront].! !

!SystemWindow methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/29/2009 12:12'!
buildWindowMenu
	"Build and answer the window menu."
	
	| aMenu |
	aMenu := self theme newMenuIn: self for: self.
	aMenu add: 'close' translated action: #closeBoxHit.
	aMenu lastItem icon: self theme windowCloseForm.
	aMenu addLine.
	aMenu add: 'about' translated action: #showAbout.
	aMenu lastItem icon: MenuIcons smallHelpIcon.
	aMenu addLine.
	aMenu addLine.
	aMenu add: 'change title...' translated action: #relabel.
	aMenu addLine.
	aMenu add: 'send to back' translated action: #sendToBack.
	aMenu add: 'make next-to-topmost' translated action: #makeSecondTopmost.
	aMenu addLine.
	self mustNotClose
		ifFalse:
			[aMenu add: 'make unclosable' translated action: #makeUnclosable]
		ifTrue:
			[aMenu add: 'make closable' translated action: #makeClosable].
	aMenu
		add: (self isSticky ifTrue: ['make draggable'] ifFalse: ['make undraggable']) translated 
		action: #toggleStickiness.
	aMenu addLine.
	self isMaximized
		ifTrue: [aMenu add: 'restore' translated action: #expandBoxHit.
				aMenu lastItem icon: self theme windowMaximizeForm]
		ifFalse: [aMenu add: 'maximize' translated action: #expandBoxHit.
				aMenu lastItem icon: self theme windowMaximizeForm].
	self isCollapsed ifFalse: [aMenu add: 'window color...' translated action: #setWindowColor].
	aMenu addLine.
	self isExclusive
		ifTrue: [aMenu add: 'make shared' translated action: #makeShared]
		ifFalse: [aMenu add: 'make exclusive' translated action: #makeExclusive].
	^aMenu! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/29/2009 12:29'!
handlesMouseOver: anEvent
	"Answer true, otherwise what is all that
	#mouseEnter:/#mouseLeave: stuff about?"

	^true! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'mig 6/3/2009 10:16'!
isExclusive

	^ self ownerHand notNil! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/29/2009 12:31'!
makeExclusive

	self ownerHand: ActiveHand.! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/29/2009 12:30'!
makeShared

	self ownerHand: nil.! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'mig 6/3/2009 10:14'!
mouseEnter: evt

	self isExclusive ifTrue: [
		(self ownerHand ~~ evt hand) ifTrue: [
			evt hand isXpfMultiPointerHandMorph ifTrue: [
				evt hand moveToPosition: ((self fullBounds outsetBy: 5) pointNearestTo: evt position).
				evt wasHandled: true.
				^ self]]].

	super mouseEnter: evt.
	! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/29/2009 12:30'!
ownerHand

	^ self valueOfProperty: #ownerHand! !

!SystemWindow methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/29/2009 12:30'!
ownerHand: aHandMorph

	^ self setProperty: #ownerHand toValue: aHandMorph! !

SharedPool subclass: #MultiPointerConstants
	instanceVariableNames: ''
	classVariableNames: 'EventTypeNone EventTypeMenu DeviceTypeMouse DeviceTypeKeyboard RedButtonBit ShiftKeyBit EventTypeDeviceDiscovered EventTypeDeviceRemoved BlueButtonBit EventTypeWindow EventTypeMouse EventKeyChar EventKeyDown YellowButtonBit OptionKeyBit EventTypeDragDropFiles CommandKeyBit EventTypeKeyboard EventKeyUp CtrlKeyBit'
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerConstants class methodsFor: 'pool initialization' stamp: 'SR 8/18/2009 16:20'!
initialize
	"EventSensorConstants initialize"
	RedButtonBit := 4.
	BlueButtonBit := 2.
	YellowButtonBit := 1.

	ShiftKeyBit := 1.
	CtrlKeyBit := 2.
	OptionKeyBit := 4.
	CommandKeyBit := 8.

	"Types of events"
	EventTypeNone := 0.
	EventTypeMouse := 1.
	EventTypeKeyboard := 2.
	EventTypeDragDropFiles := 3.
	EventTypeMenu := 4.
	EventTypeWindow := 5.
	EventTypeDeviceDiscovered := 6.
	EventTypeDeviceRemoved := 7.

	"Types of devices"
	DeviceTypeMouse := 1.
	DeviceTypeKeyboard := 2.

	"Press codes for keyboard events"
	EventKeyChar := 0.
	EventKeyDown := 1.
	EventKeyUp := 2.
! !

!TheWorldMenu methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 5/28/2009 16:18'!
windowsMenu
	"Build the windows menu for the world."
	
	^ self fillIn: (self menu: 'windows') from: {
		{ 'cascade windows'. {SystemWindow. #cascadeWindows}. 'Cascades all open windows.'}.
		{ 'order windows horizontally'. {SystemWindow. #orderWindowsHorizontally}. 'Orders all open windows horizontally.'}.
		{ 'order windows vertically'. {SystemWindow. #orderWindowsVertically}. 'Orders all open windows vertically.'}.
		nil.
		{ 'find window' . { #myWorld . #findWindow: }. 'Presents a list of all windows; if you choose one from the list, it becomes the active window.'}.
		{ 'find changed browsers...' . { #myWorld . #findDirtyBrowsers: }. 'Presents a list of browsers that have unsubmitted changes; if you choose one from the list, it becomes the active window.'}.
		{ 'find changed windows...' . { #myWorld . #findDirtyWindows: }. 'Presents a list of all windows that have unsubmitted changes; if you choose one from the list, it becomes the active window.'}.
		nil.
		{ 'find a transcript (t)' . { #myWorld . #findATranscript: }. 'Brings an open Transcript to the front, creating one if necessary, and makes it the active window'}.
		{ 'find a fileList (L)' . { #myWorld . #findAFileList: }. 'Brings an open fileList  to the front, creating one if necessary, and makes it the active window'}.
		{ 'find a change sorter (C)' . { #myWorld . #findAChangeSorter: }. 'Brings an open change sorter to the front, creating one if necessary, and makes it the active window'}.
		{ 'find message names (W)' . { #myWorld . #findAMessageNamesWindow: }. 'Brings an open MessageNames window to the front, creating one if necessary, and makes it the active window'}.
		nil.
		{ #staggerPolicyString . { self . #toggleWindowPolicy }. 'stagger: new windows positioned so you can see a portion of each one. tile: new windows positioned so that they do not overlap others, if possible.'}.
		nil.
		{ 'collapse all windows' . { #myWorld . #collapseAll }. 'Reduce all open windows to collapsed forms that only show titles.'}.
		{ 'expand all windows' . { #myWorld . #expandAll }. 'Expand all collapsed windows back to their expanded forms.'}.
		{ 'close top window (w)' . { SystemWindow . #closeTopWindow }. 'Close the topmost window if possible.'}.
		{ 'send top window to back (\)' . { SystemWindow . #sendTopWindowToBack  }. 'Make the topmost window become the backmost one, and activate the window just beneath it.'}.
		{ 'move windows onscreen' . { #myWorld . #bringWindowsFullOnscreen }. 'Make all windows fully visible on the screen'}.
		nil.
		{ 'delete unchanged windows' . { #myWorld . #closeUnchangedWindows }. 'Deletes all windows that do not have unsaved text edits.'}.
		{ 'delete non-windows' . { #myWorld . #deleteNonWindows }. 'Deletes all non-window morphs lying on the world.'}.
		{ 'delete both of the above' . { self . #cleanUpWorld }. 'deletes all unchanged windows and also all non-window morphs lying on the world, other than flaps.'}.
	}! !

!MenuItemMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/22/2009 18:28'!
mouseDown: evt
	"Handle a mouse down event.
	Menu items get activated when the mouse is over them."

	"enable label editing" 
	evt shiftPressed ifTrue: [^ super mouseDown: evt].  

	"Bring menu to the front if it is obscured."
	self owner owner = self world ifTrue: [
		self owner comeToFront].

	(self isInDockingBar
			and:[isSelected])
		ifTrue:[
			evt hand newMouseFocus: nil.
			owner selectItem: nil event: evt. ]
		ifFalse:[
			"Redirect to menu for valid transitions"
			evt hand newMouseFocus: owner. 
			owner selectItem: self event: evt. ]
! !

!NewParagraph methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 6/26/2009 12:39'!
displaySelectionInLine: line on: aCanvas 

	| leftX rightX w caretColor |
	aCanvas isShadowDrawing ifTrue: [ ^self ].	"don't draw selection with shadow"
		
	World handsDo: [:hand | | currentSelectionStart currentSelectionStop|
		currentSelectionStart := self selectionStartForHand: hand.
		currentSelectionStop := self selectionStopForHand: hand.
		(self showSelectionFrom: currentSelectionStart to: currentSelectionStop in: line) ifTrue: [
			leftX := (currentSelectionStart stringIndex < line first 
				ifTrue: [line]
				ifFalse: [currentSelectionStart ]) left.
			rightX := (currentSelectionStop stringIndex > (line last + 1)
				or:  [currentSelectionStop stringIndex = (line last + 1) 
				and: [currentSelectionStop textLine ~= line]]) 
				ifTrue: [line right]
				ifFalse: [currentSelectionStop left].
			currentSelectionStart = currentSelectionStop 
				ifTrue: [
					rightX := rightX + 1.
					w := self caretWidth.
					caretColor := self insertionPointColor.
					1 to: w do:  [:i | 
						"Draw caret triangles at top and bottom"
						aCanvas
							fillRectangle: ((leftX - w + i - 1) @ (line top + i - 1) 
							extent: ((w - i) * 2 + 3) @ 1)
							color: caretColor.
						aCanvas
							fillRectangle: ((leftX - w + i - 1) @ (line bottom - i) 
							extent: ((w - i) * 2 + 3) @ 1)
							color: caretColor].
					aCanvas
						fillRectangle: (leftX @ line top corner: rightX @ line bottom)
						color: caretColor]
				ifFalse: [
					aCanvas
						fillRectangle: (leftX @ line top corner: rightX @ line bottom)
						color: ((self selectionColorForHand: hand) alpha: 0.2)]]].! !

!NewParagraph methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/6/2009 01:29'!
selectionColorForHand: aHand

	| color |
	aHand isXpfMultiPointerHandMorph
		ifTrue: [^ aHand handColor]
		ifFalse: [
			Display depth = 1 ifTrue: [^ Color veryLightGray].
			Display depth = 2 ifTrue: [^ Color gray].
			color := Preferences textHighlightColor.
			self focused ifFalse: [color := color alphaMixed: 0.2 with: Color veryVeryLightGray].
			^ color].! !

!NewParagraph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 01:30'!
selectionRects
	"Return an array of rectangles representing the selection region."
	| rects |
	selectionStart isEmptyOrNil ifTrue: [^ Array new].
	
	rects := OrderedCollection new.
	World handsDo: [:hand |
		(self selectionStartForHand: hand) isNil ifFalse: [
			rects addAll: (self selectionRectsFrom: (self selectionStartForHand: hand) to: (self selectionStopForHand: hand))]].
	^ rects.! !

!NewParagraph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 14:48'!
selectionStart

	^ self selectionStartForHand: ActiveHand! !

!NewParagraph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 14:47'!
selectionStart: startBlock selectionStop: stopBlock

	^ self selectionStart: startBlock selectionStop: stopBlock forHand: ActiveHand! !

!NewParagraph methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/6/2009 14:46'!
selectionStart: startBlock selectionStop: stopBlock forHand: aHand

	selectionStart ifNil: [
		selectionStart := Dictionary new].
	selectionStart isDictionary ifFalse: [
		selectionStart := Dictionary new].
	selectionStart at: aHand put: startBlock.
	
	selectionStop ifNil: [
		selectionStop := Dictionary new].
	selectionStop isDictionary ifFalse: [
		selectionStop := Dictionary new].
	selectionStop at: aHand put: stopBlock.! !

!NewParagraph methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/6/2009 01:29'!
selectionStartForHand: aHand

	^ selectionStart isDictionary
		ifTrue: [selectionStart at: aHand ifAbsent: [nil]]
		ifFalse: [selectionStart]! !

!NewParagraph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/6/2009 14:47'!
selectionStop

	^ self selectionStartForHand: ActiveHand! !

!NewParagraph methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/6/2009 01:29'!
selectionStopForHand: aHand

	^ selectionStop isDictionary
		ifTrue: [selectionStop at: aHand ifAbsent: [nil]]
		ifFalse: [selectionStop]! !

!NewParagraph methodsFor: '*xpf-multipoint-core' stamp: 'SR 6/6/2009 00:53'!
showSelectionFrom: SelectionStart to: SelectionStop in: line
	"Check wether a selection should be displayed"
	SelectionStart ifNil: [^ false].
	SelectionStart = SelectionStop
		ifTrue:  [
			"Only show caret on line where clicked"
			SelectionStart textLine ~= line ifTrue: [^ false]]
		ifFalse:  [
			"Test entire selection before or after here"
			(SelectionStop stringIndex < line first 
				or: [SelectionStart stringIndex > (line last + 1)]) ifTrue: [^ false].	"No selection on this line"
			(SelectionStop stringIndex = line first 
				and: [SelectionStop textLine ~= line]) ifTrue: [^ false].	"Selection ends on line above"
			(SelectionStart stringIndex = (line last + 1) 
				and: [SelectionStop textLine ~= line]) ifTrue: [^ false]].	"Selection begins on line below"
	^ true! !

!MenuMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 6/20/2009 19:06'!
popUpAt: aPoint forHand: hand in: aWorld allowKeyboard: aBoolean 
	"Present this menu at the given point under control of the given  
	hand."
	| evt |
 	MultiPointerManager multiPointerActive ifFalse: [
 		aWorld submorphs
 			select: [ :each | (each isKindOf: MenuMorph)
 				and: [each stayUp not]]
 			thenDo: [ :menu | menu delete]].

	self items isEmpty
		ifTrue: [^ self].

	MenuIcons decorateMenu: self.

	(self submorphs
		select: [:m | m isKindOf: UpdatingMenuItemMorph])
		do: [:m | m updateContents].
	"precompute width"
	self
		positionAt: aPoint
		relativeTo: (selectedItem
				ifNil: [self items first])
		inWorld: aWorld.
	aWorld addMorphFront: self.
	"Acquire focus for valid pop up behavior"
	hand newMouseFocus: self.
	aBoolean
		ifTrue: [hand newKeyboardFocus: self].
	evt := hand lastEvent.
	(evt isKeyboard
			or: [evt isMouse
					and: [evt anyButtonPressed not]])
		ifTrue: ["Select first item if button not down"
			self moveSelectionDown: 1 event: evt].
	self updateColor.
	self changed! !

Object subclass: #MultiPointer
	instanceVariableNames: 'pointerID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointer class
	instanceVariableNames: 'bufferSemaphore'!

!MultiPointer commentStamp: 'mig 5/19/2009 19:49' prior: 0!
This class provides abstractions to the MultiPointer plugin. Instances are bound to a specific pointer in the operating system and provide messages for retrieving information about them.!

MultiPointer class
	instanceVariableNames: 'bufferSemaphore'!

!MultiPointer class methodsFor: 'instance creation' stamp: 'SR 8/19/2009 13:57'!
allPointers

	^ {}
	"^ (0 to: self getPointerCount - 1) collect: [:id | self newForID: id]"! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 8/13/2009 14:28'!
bufferSemaphore

	^ bufferSemaphore ifNil: [self bufferSemaphore: Semaphore new. bufferSemaphore]! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 8/13/2009 14:11'!
bufferSemaphore: anObject

	bufferSemaphore := anObject.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/19/2009 13:52'!
eventAvailable

	^ self pvtEventAvailable ~= 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:21'!
getDeviceCount

	<primitive: 'primitiveMultiPointerDeviceCount' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:29'!
getDeviceLively: aPointerID

	^ (self pvtGetDeviceLively: aPointerID) ~= 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:57'!
getNextEvent: array

	<primitive: 'primitiveMultiPointerGetNextEvent' module: 'MultiPointerPlugin'>
	"self primitiveFailed."

! !

!MultiPointer class methodsFor: 'instance creation' stamp: 'mig 5/19/2009 10:51'!
newForID: aPointerID

	^ self new pointerID: aPointerID! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/19/2009 13:52'!
pvtEventAvailable

	<primitive: 'primitiveMultiPointerEventAvailable' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:18'!
pvtGetDeviceLively: aPointerID

	<primitive: 'primitiveMultiPointerDeviceLively' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 8/13/2009 14:20'!
registerSemaphores

	self setBufferSemaphoreHandle: (Smalltalk registerExternalObject: self bufferSemaphore).! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/19/2009 14:58'!
resetAll

	self allPointers do: [:pointer | pointer isLively ifTrue: [pointer reset]].! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/27/2009 15:02'!
resetPlugin

	<primitive: 'primitiveMultiPointerReset' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:19'!
setBufferSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerSetBufferSemaphore' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 8/13/2009 14:21'!
unregisterSemaphores

	self setBufferSemaphoreHandle: 0.
	Smalltalk unregisterExternalObject: self bufferSemaphore.
	self bufferSemaphore: nil.
! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:20'!
useMultiInput

	<primitive: 'primitiveMultiPointerUseMultiInput' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:20'!
useOSInput

	<primitive: 'primitiveMultiPointerUseOSInput' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

Object subclass: #MultiPointerManager
	instanceVariableNames: 'handMorphs colors queueProcessor moving handMorphLookingForKeyboard'
	classVariableNames: ''
	poolDictionaries: 'MultiPointerConstants'
	category: 'Xpf-Multipoint-Core'!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager commentStamp: 'mig 5/19/2009 19:47' prior: 0!
A MultiPointerManager controls all MultiPointer actions. There should be only one MultiPointerManager instance in a running system. Instances are created and deleted upon startup and shutdown.

colors: is a dictionary having colors as keys and usage counts as values.
deviceUpdateProcess and movementProcess: hold the two forked processes for MultiPointer.
handMorphs: is a list of all multipointer hands.
keyboardHand: is the hand which sends keyboard events into the world.
!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager class methodsFor: 'class initialization' stamp: 'mig 5/19/2009 11:44'!
initialize

	super initialize.
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.
	! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance

	^ instance! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance: anObject

	instance := anObject.! !

!MultiPointerManager class methodsFor: 'as yet unclassified' stamp: 'SR 8/27/2009 21:43'!
keyboardWaitTime

	^ 60! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:37'!
multiPointerActive

	^ self instance notNil and: [self instance multiPointerActive]! !

!MultiPointerManager class methodsFor: 'testing' stamp: 'SR 5/6/2009 10:19'!
pluginAvailable

	^ SmalltalkImage current listBuiltinModules anySatisfy: [:each | each startsWith: 'MultiPointerPlugin']! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:55'!
shutDown: aboutToQuit 

	self instance ifNotNil: [
		self instance shutdown.
		self instance: nil].
	
	super shutDown: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
startMoving

	self instance ifNotNil: [ self instance startMoving ]! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:54'!
startUp: aboutToQuit 

	self pluginAvailable ifTrue: [
		self instance: self new.
		self instance startup].
	
	super startUp: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
stopMoving

	self instance ifNotNil: [ self instance stopMoving ]! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/18/2009 15:32'!
addAllHands

	self handMorphs do: [:hand | 	
		World addHand: hand.
		hand changed].
! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/27/2009 21:55'!
assignKeyboardToLookingHand: keyboardId 

	self handMorphLookingForKeyboard keyboardIndex ifNotNil: [^ self].
	(self handMorphs anySatisfy: [:morph | morph keyboardIndex = keyboardId]) ifTrue: [^ self].
	
	self handMorphLookingForKeyboard keyboardIndex: keyboardId.
	self handMorphLookingForKeyboard: nil.! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors

	^ colors! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors: anObject

	colors := anObject.! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 15:25'!
createHandMorphFor: aMouseIndex
	
	| hand |
	(self handMorphForMouseID: aMouseIndex) ifNotNil: [^ self].

	hand := MultiPointerHandMorph new 
		mouseIndex: aMouseIndex;
		handColor: self getColor.
	self handMorphs add: hand.
	self multiPointerActive ifTrue: [World addHand: hand].
! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 13:23'!
destroyHandMorphFor: aMouseIndex

	(self handMorphs detect: [:each | each mouseIndex = aMouseIndex] ifNone: [nil])
		ifNotNilDo: [:morph |
			self handMorphs remove: morph.
			World removeHand: morph.
			self releaseColor: morph handColor].
! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 21:56'!
findKeyboardForHandMorph: aHandMorph

	self handMorphLookingForKeyboard ifNotNil: [^ self].
	aHandMorph isXpfMultiPointerHandMorph ifFalse: [^ self].
	aHandMorph keyboardIndex ifNotNil: [^ self].
	
	self handMorphLookingForKeyboard: aHandMorph.
	(1 to: self class keyboardWaitTime)
		do: [:val | 
			aHandMorph keyboardIndex ifNotNil: [^ self].
			(Delay forSeconds: 1) wait]
		displayingProgress: 'Waiting for key press'! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:26'!
getColor

	| color |
	color := self colors keyAtValue: self colors values min.
	self colors at: color put: (self colors at: color) + 1.
	^ color! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 15:24'!
handMorphForKeyboardID: aKeyboardIndex
	
	^ self handMorphs detect: [:each | each keyboardIndex = aKeyboardIndex] ifNone: [nil]
! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 15:24'!
handMorphForMouseID: aMouseIndex
	
	^ self handMorphs detect: [:each | each mouseIndex = aMouseIndex] ifNone: [nil]
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/27/2009 21:44'!
handMorphLookingForKeyboard

	^ handMorphLookingForKeyboard! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/27/2009 21:45'!
handMorphLookingForKeyboard: anObject

	handMorphLookingForKeyboard := anObject! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs

	^ handMorphs! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs: anObject

	handMorphs := anObject.! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/27/2009 13:23'!
handleDetectedEvent: evt

	| type index |
	type := evt at: 3.
	index := evt at: 4.
	type = DeviceTypeMouse ifTrue: [self createHandMorphFor: index].! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/27/2009 13:23'!
handleRemovedEvent: evt

	| type index |
	type := evt at: 3.
	index := evt at: 4.
	type = DeviceTypeMouse ifTrue: [self destroyHandMorphFor: index].! !

!MultiPointerManager methodsFor: 'initialize-release' stamp: 'SR 8/18/2009 15:38'!
initialize

	super initialize.
	
	self
		moving: false;
		handMorphs: OrderedCollection new;
		initializeColors.! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:44'!
initializeColors

	self colors: (Dictionary newFrom: {
		Color red -> 0.
		Color orange -> 0.
		Color green -> 0.
		Color blue -> 0.
	}).
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:38'!
moving

	^ moving! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:38'!
moving: anObject

	moving := anObject! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
multiPointerActive

	^ self queueProcessor notNil! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
queueProcessor

	^ queueProcessor! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
queueProcessor: anObject

	queueProcessor := anObject! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/27/2009 22:31'!
queueProcessorBlock

	^ [[
		MultiPointer bufferSemaphore wait.
		[MultiPointer eventAvailable] whileTrue: [ | eventBuffer type |
			eventBuffer := Array new: 8.
			MultiPointer getNextEvent: eventBuffer.
			type := eventBuffer at: 1.
			"Transcript show: eventBuffer;cr."
		
			type = EventTypeDeviceDiscovered ifTrue: [self handleDetectedEvent: eventBuffer].
			type = EventTypeDeviceRemoved ifTrue: [self handleRemovedEvent: eventBuffer].
			self moving ifTrue: [self sendHandMorphEvent: eventBuffer].
			(type = EventTypeKeyboard and: [self handMorphLookingForKeyboard isNil not]) ifTrue: [self assignKeyboardToLookingHand: (eventBuffer at: 7)]].
	  ] repeat ]! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:28'!
releaseColor: aColor

	self colors at: aColor put: (self colors at: aColor) - 1! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/27/2009 14:10'!
removeAllHands

	self handMorphs do: [:each | 
		World removeHand: each].
! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/27/2009 20:51'!
sendHandMorphEvent: eventBuffer

	| type index handMorph |
	type := eventBuffer at: 1.
	index := eventBuffer at: 7.
	
	handMorph := nil.
	type = EventTypeMouse ifTrue: [handMorph := self handMorphForMouseID: index].
	type = EventTypeKeyboard ifTrue: [handMorph := self handMorphForKeyboardID: index].
	handMorph ifNil: [^ self].
	
	handMorph enqueueEvent: eventBuffer.! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'SR 8/18/2009 15:46'!
shutdown

	self stopMoving.
	self stopQueueProcessor.
	MultiPointer unregisterSemaphores.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/27/2009 22:05'!
startMoving

	self moving ifTrue: [^ self].
	
	MultiPointer useMultiInput.
	self
		moving: true;
		addAllHands.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 9/1/2009 17:25'!
startQueueProcessor

	self queueProcessor ifNotNil: [^self].
	
	self queueProcessor: 
		(self queueProcessorBlock
			forkAt: Processor userSchedulingPriority
			named: 'MultiPointer Queue Processor').! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'SR 8/27/2009 15:14'!
startup

	MultiPointer resetPlugin.
	MultiPointer registerSemaphores.
	self startQueueProcessor.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/27/2009 22:06'!
stopMoving

	self moving ifFalse: [^ self].
	
	MultiPointer useOSInput.
	self 
		moving: false;
		removeAllHands.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/18/2009 15:42'!
stopQueueProcessor

	self queueProcessor ifNotNil: [
		self queueProcessor terminate.
		self queueProcessor: nil.
		
		MultiPointer allPointers do: [:each |
			self destroyHandMorphFor: each].
		
		self handMorphs: OrderedCollection new].
! !

!Cursor class methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 5/18/2009 15:03'!
extent: extentPoint fromArray: anArray offset: offsetPoint 
	"Answer a new instance of me with width and height specified by
	extentPoint, offset by offsetPoint, and bits from anArray.
	The width must be smaller or equal to 32 to fit into an 32-bit bitmap"
	
	(extentPoint x between: 1 and: 32) ifFalse: [
		^ self error: 'cursors must be max. 32 pixels wide'].

	^ super
		extent: extentPoint
		fromArray: (anArray collect: [:bits | bits bitShift: 32 - extentPoint x])
		offset: offsetPoint.
! !

!SharedQueue methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/4/2009 10:34'!
nextPut: value 
	"Send value through the receiver. If a Process has been suspended 
	waiting to receive a value through the receiver, allow it to proceed."

	value ifNil: [^ self].
	accessProtect
		critical: [writePosition > contentsArray size
						ifTrue: [self makeRoomAtEnd].
				 contentsArray at: writePosition put: value.
				 writePosition := writePosition + 1].
	readSynch signal.
	^value! !

!SharedQueue methodsFor: '*xpf-multipoint-core' stamp: 'SR 4/29/2009 17:24'!
nextPutAll: aCollection

	aCollection do: [:each | self nextPut: each].! !

!Morph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/22/2009 18:27'!
delete
	"Remove the receiver as a submorph of its owner and make its 
	new owner be nil."

	| aWorld |
	self removeHalo.
	aWorld := self world ifNil: [World].
	"Terminate genie recognition focus"
	"I encountered a case where the hand was nil, so I put in a little 
	protection - raa "
	" This happens when we are in an MVC project and open
	  a morphic window. - BG "
	aWorld ifNotNil: [
		aWorld handsDo:[:hand |
			self disableSubmorphFocusForHand: hand.
			hand
				releaseKeyboardFocus: self;
				releaseMouseFocus: self]].
	owner ifNotNil:[ self privateDelete.
		self player ifNotNilDo: [ :player |
			"Player must be notified"
			player noteDeletionOf: self fromWorld: aWorld]].! !
MultiPointerConstants initialize!
MultiPointerManager initialize!
